@@ -51,7 +51,7 @@ impl SaveFile {
             save_version: file.read_i32::<L>()?,
             build_version: file.read_i32::<L>()?,
             world_type: read_string(file)?,
-            world_properties: WorldProperties::new(&read_string(file)?)?,
+            world_properties: WorldProperties::parse(&read_string(file)?)?,
             session_name: read_string(file)?,
             play_time: Duration::seconds(file.read_i32::<L>()?.try_into()?),
             save_date: SaveFile::convert_date(file.read_i64::<L>()?),
@@ -110,7 +110,7 @@ pub struct WorldProperties {
 }
 
 impl WorldProperties {
-    pub fn new(s: &str) -> Result<WorldProperties> {
+    pub fn parse(s: &str) -> Result<WorldProperties> {
         let mut map: HashMap<&str, &str> = s
             .split('?')
             .skip(1) // Nothing before first "?"
@@ -218,25 +218,39 @@ impl SaveObject {
     }
 }
 
-fn read_string<R>(file: &mut R) -> Result<String>
+pub fn read_string<R>(file: &mut R) -> Result<String>
 where
     R: Read,
 {
-    let length = file.read_i32::<L>()?;
+    const MAX_LENGTH: usize = 0x1000;
+    let length_error = || Error::msg("invalid length");
+    let signed_length = file.read_i32::<L>()?;
+
+    Ok(if signed_length < 0 {
+        // Negation fails with minimum i32
+        if signed_length == i32::MIN {
+            return Err(length_error());
+        }
 
-    Ok(if length < 0 {
         let mut buffer: Vec<u16> = Vec::new();
-        buffer.resize(((-length) as usize).saturating_sub(1) / 2, 0);
+        let length = ((-signed_length) as usize).saturating_sub(1) / 2;
+        if length > MAX_LENGTH {
+            return Err(length_error());
+        }
+        buffer.resize(length, 0);
         file.read_u16_into::<L>(&mut buffer)?;
         String::from_utf16_lossy(&buffer)
     } else {
         let mut buffer: Vec<u8> = Vec::new();
-        buffer.resize((length.abs() as usize).saturating_sub(1), b'\0');
+        let length = (signed_length as usize).saturating_sub(1);
+        if length > MAX_LENGTH {
+            return Err(length_error());
+        }
+        buffer.resize(length, b'\0');
         file.read_exact(&mut buffer)?;
         if length > 0 {
             // Skip null char
-            let b = file.read_u8()?;
-            debug_assert_eq!(b, b'\0');
+            file.read_u8()?;
         }
         String::from_utf8_lossy(&buffer).into_owned()
     })
@@ -335,9 +349,9 @@ mod tests {
 
     #[test]
     fn world_properties() {
-        assert!(WorldProperties::new("").is_err());
+        assert!(WorldProperties::parse("").is_err());
         let string = "?startloc=Grass Fields?sessionName=test_file?Visibility=SV_Private";
-        let result = WorldProperties::new(string).unwrap();
+        let result = WorldProperties::parse(string).unwrap();
         assert_eq!(result.start_loc, "Grass Fields");
         assert_eq!(result.session_name, "test_file");
         assert_eq!(result.visibility, SessionVisiblity::SvPrivate);
@@ -357,14 +371,24 @@ mod tests {
     fn test_read_string() {
         {
             // Empty file
-            let mut data = Cursor::new(Vec::new());
-            assert!(read_string(&mut data).is_err());
+            assert!(read_string(&mut &Vec::new()[..]).is_err());
         }
         {
             // Just the prefix
             let mut data = &0_i32.to_le_bytes()[..];
             assert_eq!(read_string(&mut data).unwrap(), "");
         }
+
+        // Invalid lengths
+        let cases: &[&[u8]] = &[
+            &i32::MIN.to_le_bytes()[..],
+            &(i32::MIN + 1).to_le_bytes()[..],
+            &i32::MAX.to_le_bytes()[..],
+        ];
+        for data in cases {
+            assert!(read_string(&mut &data.to_vec()[..]).is_err());
+        }
+
         // Various strings
         for test_string in &["", "a", "abc"] {
             let encoded = to_encoding(test_string.as_bytes());
